#   Copyright 2022 Stian Soiland-Reyes
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""
Parse HTTP headers to find Signposting links
"""

from typing import Dict, List, Set, Tuple, Optional, Collection, Set
import httplink
from httplink import ParsedLinks, Link, parse_link_header
from urllib.parse import urljoin

# Only relations listed below will be selected
# Sources:
#   https://signposting.org/conventions/
#   https://signposting.org/FAIR/
"""Valid Signposting link relations"""
SIGNPOSTING=set("author collection describedby describes item cite-as type license linkset".split(" "))

def _filter_links_by_rel(parsedLinks:ParsedLinks, *rels:str) -> List[Link]:
    """Filter links to select from a set of relations.

    The relations must be valid Signposting relation listed in `SIGNPOSTING`.

    Return a list of ``Link``s which ``rel`` match the given ``rels``.
    
    """
    if rels:        
        # Ensure all filters are in lower case
        filterRels = set(r.lower() for r in rels)
        unknown = filterRels - SIGNPOSTING
        if unknown:
            raise ValueError("Unknown FAIR Signposting relations: %s" % unknown)
    else:
        # Fallback - all valid signposting relations
        filterRels = SIGNPOSTING
    return [l for l in parsedLinks.links if l.rel & filterRels]

def _optional_link(parsedLinks:ParsedLinks, rel:str) -> Optional[Link]:
    """Look up a single link relation.

    The relation must be a valid Signposting relation listed in `SIGNPOSTING`.

    It is undefined which relation is returned if multiple links of the same
    relation are found.

    Return the ``Link`` or ``None`` if not found.
    """
    if not rel.lower() in SIGNPOSTING:
        raise ValueError("Unknown FAIR Signposting relation: %s" % rel)
    if rel in parsedLinks:
        return parsedLinks[rel]
    return None


class Signposting:
    """Signposting links for a given resource.
    
    Links are discovered according to `FAIR`_ `signposting`_ conventions.

    .. _signposting: https://signposting.org/conventions/
    .. _FAIR: https://signposting.org/FAIR/
    """

    """Resource URL this is the signposting for, e.g. a HTML landing page.
    
    Note that individual links may override the `context`_ using the ``Link["anchor"]`` attribute.

    .. _context: https://datatracker.ietf.org/doc/html/rfc8288#section-3.2

    """
    context_url: Optional[str]

    """Author(s) of the resource (and presuambly it items)"""
    author: List[Link]

    """Metadata resources about the resource and its items, typically in a Linked Data format. 
    
    Resources may require content negotiation, check ``Link["type"]`` attribute
    (if present) for content type, e.g. ``text/turtle``.
    """
    describedBy: List[Link]

    """Semantic types of the resource, e.g. from schema.org"""
    type: List[Link]

    """Items contained by this resource, e.g. downloads.
    
    The content type of the download may be available as ``Link["type"]``` attribute.
    """
    item: List[Link]

    """Linkset resuorces with further signposting.

    A `linkset`_ is a JSON or text serialization of Link headers available as a
    separate resource, and may be used to externalize large collection of links, e.g.
    thousands of "item" relations.

    Resources may require content negotiation, check ``Link["type"]`` attribute
    (if present)  for content types ``application/linkset`` or ``application/linkset+json``.

    .. _linkset: https://datatracker.ietf.org/doc/draft-ietf-httpapi-linkset/
    """
    linkset: List[Link]

    """Persistent Identifier (PID) for this resource, preferred for citation and permalinks"""
    citeAs: Optional[Link]

    """Optional license of this resource (and presumably its items)"""
    license: Optional[Link]

    """Optional collection this resource is part of"""
    collection: Optional[Link]

    def __init__(self, parsedLinks:ParsedLinks, context_url:str=None):
        """Initialize Signposting object for a given `ParsedLinks` 
        as discovered from the (optional) `context_url` base URL.
        """
        # According to FAIR Signposting
        # <https://www.signposting.org/FAIR/> version 20220225
        self.context_url = context_url
        self.author = _filter_links_by_rel(parsedLinks, "author")
        self.describedBy = _filter_links_by_rel(parsedLinks, "describedby")
        self.type = _filter_links_by_rel(parsedLinks, "type")
        self.item = _filter_links_by_rel(parsedLinks, "item")
        self.linkset =  _filter_links_by_rel(parsedLinks, "linkset")
        self.citeAs = _optional_link(parsedLinks, "cite-as")
        self.license = _optional_link(parsedLinks, "license")
        self.collection = _optional_link(parsedLinks, "collection")

def _absolute_attribute(k:str, v:str, baseurl:str) -> Tuple[str,str]:
    """Ensure link attribute value uses absolute URI, resolving from the baseurl.

    Currently this will mean the `context`_ attribute ``anchor`` will be rewritten.

    .. _context: https://datatracker.ietf.org/doc/html/rfc8288#section-3.2
    """
    if k.lower() == "anchor":
        return k, urljoin(baseurl, v)
    return k, v

def find_signposting(headers:List[str], baseurl:str=None) -> Signposting:
    """Find signposting among HTTP Link headers. 

    Optionally make the links absolute according to the base URL.

    The link headers should be valid according to `RFC8288`_, excluding the "Link:" prefix.
    
    Links are discovered according to defined `FAIR`_ `signposting`_ relations.

    .. _signposting: https://signposting.org/conventions/
    .. _FAIR: https://signposting.org/FAIR/
    .. _rfc8288: https://doi.org/10.17487/RFC8288
    
    """
    parsed = parse_link_header(", ".join(headers))
    signposting: List[Link] = []
    # Ignore non-Signposting relations like "stylesheet"
    for l in _filter_links_by_rel(parsed):
        if baseurl is not None:
            # Make URLs absolute by modifying Link object in-place
            target = urljoin(baseurl, l.target)
            attributes = [_absolute_attribute(k,v, baseurl) for k,v in l.attributes]            
            link = Link(target, attributes)
        else:
            link = l # unchanged
        signposting.append(link)
    return Signposting(ParsedLinks(signposting), baseurl)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
"""Setup dot py."""
from __future__ import absolute_import, print_function

# import re
from glob import glob
from os.path import basename, dirname, join, splitext

from setuptools import find_packages, setup


def read(*names, **kwargs):
    """Read description files."""
    path = join(dirname(__file__), *names)
    with open(path, encoding=kwargs.get('encoding', 'utf8')) as fh:
        return fh.read()


# previous approach used to ignored badges in PyPI long description
# long_description = '{}\n{}'.format(
#     re.compile(
#         '^.. start-badges.*^.. end-badges',
#         re.M | re.S,
#         ).sub(
#             '',
#             read('README.rst'),
#             ),
#     re.sub(':[a-z]+:`~?(.*?)`', r'``\1``', read(join('docs', 'CHANGELOG.rst')))
#     )

long_description = '{}\n{}'.format(
    read('README.rst'),
    read(join('docs', 'CHANGELOG.rst')),
    )

setup(
    name='signposting',
    version='0.1.0',
    description='Parse and navigate FAIR Signposting Link headers',
    long_description=long_description,
    long_description_content_type='text/x-rst',
    license='Apache License, version 2.0',
    author='Stian Soiland-Reyes',
    author_email='stain@apache.org',
    url='https://github.com/stain/signposting',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    py_modules=[splitext(basename(i))[0] for i in glob("src/*.py")],
    include_package_data=True,
    zip_safe=True,
    classifiers=[
        # complete classifier list:
        # http://pypi.python.org/pypi?%3Aaction=list_classifiers
        'Development Status :: 1 - Planning',
        'Environment :: Web Environment',
        'License :: OSI Approved :: Apache Software License',
        'Intended Audience :: Science/Research',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Topic :: Internet :: WWW/HTTP'
        ],
    project_urls={
        'webpage': 'https://github.com/stain/signposting',
        #'Documentation': 'https://signposting.readthedocs.io/en/latest/',
        'Changelog': 'https://github.com/stain/signposting/blob/main/docs/CHANGELOG.rst',
        'Issue Tracker': 'https://github.com/stain/signposting/issues',
        'Discussion Forum': 'https://github.com/stain/signposting/discussions',
        },
    keywords=[
        'FAIR', 'signposting', 'linked data',
        'DOI', 'HTTP', 'linkset'
        ],
    python_requires='>=3.6',
    install_requires=[
        'beautifulsoup4>=4.10',
        'httplink==0.2.0',
        ],
    extras_require={
        # eg:
        #   'rst': ['docutils>=0.11'],
        #   ':python_version=="2.6"': ['argparse'],
        },
    setup_requires=[
        #   'pytest-runner',
        #   'setuptools_scm>=3.3.1',
        ],
    entry_points={
        'console_scripts': [
            'signposting= signposting.cli:main',
            ]        
        },
    # cmdclass={'build_ext': optional_build_ext},
    # ext_modules=[
    #    Extension(
    #        splitext(relpath(path, 'src').replace(os.sep, '.'))[0],
    #        sources=[path],
    #        include_dirs=[dirname(path)]
    #    )
    #    for root, _, _ in os.walk('src')
    #    for path in glob(join(root, '*.c'))
    # ],
    )



"""
DOCSTRING for first public console interface.

USAGE:
    signposting
"""

from typing import Dict, List, Set, Tuple, Optional, Collection, Set
import argparse
import sys
import enum
from urllib.error import HTTPError, URLError

from . import Signposting,find_signposting,find_signposting_http,Link

def _multiline(header:str, lines:Collection[str]):
    indent = "\n" + (" " * (len(header)+2))
    return "%s: %s" % (header, indent.join(lines))

def _target(link: Link):
    return "<%s>" % link.target

def _target_and_type(link: Link):
    return "<%s> %s" % (link.target, 
                      "type" in link and link["type"] or "")

errors = enum.IntEnum("Error", 
    "OK BAD_URL HTTP_ERROR LINK_SYNTAX INTERNAL_ERROR",
    start=0
)

def main(*args:str):
    """Discover signposting"""
    
    parser = argparse.ArgumentParser()
    parser.add_argument("url", nargs='+', 
        help="URL(s) to discover signposting for")
    if args:
        parsed = parser.parse_args(args)
    else:        
        parsed = parser.parse_args()
    isFirst = True
    for url in parsed.url:
        if not isFirst:
            isFirst = False
        else:
            print() ## separator

        try:
            signposting = find_signposting_http(url)
        except URLError as e:
            print("Failed URL %s" % url, file=sys.stderr)
            print("%s" % e.reason, file=sys.stderr)
            return errors.BAD_URL
        except HTTPError as e:
            print("HTTP error for %s" % url, file=sys.stderr)
            print("%s" % e.reason, file=sys.stderr)
            return errors.HTTP_ERROR
        except ValueError as e:
            print("Could not parse Link header for %s" % url, file=sys.stderr)
            print("%s" % e, file=sys.stderr)
            return errors.LINK_SYNTAX
    #    except Exception as e:
    #        print("%s" % e, file=sys.stderr)
    #        return errors.INTERNAL_ERROR

        print("Signposting for", signposting.context_url or url)
        if (signposting.citeAs):            
            print("CiteAs:", _target(signposting.citeAs))
        if (signposting.type):
            print(_multiline("Type", [_target(l) for l in signposting.type]))
        if (signposting.collection):
            print("Collection:", _target(signposting.collection))
        if (signposting.license):
            print("License:", _target(signposting.license))
        if (signposting.author):
            print(_multiline("Author", [_target(l) for l in signposting.author]))
        if (signposting.describedBy):
            print(_multiline("DescribedBy", [_target_and_type(l) for l in signposting.describedBy]))
        if (signposting.item):
            print(_multiline("Item", [_target_and_type(l) for l in signposting.item]))
        if (signposting.linkset):
            print(_multiline("Linkset", [_target_and_type(l) for l in signposting.linkset]))
    return errors.OK


"""Finding signposting in FAIR resources.

This library helps client to discover links that follow the 
`FAIR`_ `signposting`_ conventions.

This can then be used to navigate between:

* Persistent identifiers
* HTML landing pages
* File downloads/items
* Structured metadata

The library works by inspecting the HTTP messages for
``Link`` headers from a given URI with `find_signposting_http`, which
which categorize them by their `rel` `Link relation`_ into a 
`Signposting` object with absolute URIs.

It is up to the clients of this library to decide how to further
navigate or retrieve the associated resources, e.g. using a 
RDF library like `rdflib`_.

Future versions of this library may also provide ways to discover
FAIR signposting in HTML ``<link>`` annotations and in 
`linkset`_ documents.

.. _signposting: https://signposting.org/conventions/
.. _FAIR: https://signposting.org/FAIR/
.. _Link Relation: https://www.iana.org/assignments/link-relations/
.. _rdflib: https://rdflib.readthedocs.io/en/stable/
.. _linkset: https://signposting.org/FAIR/#linksetrec
"""

__version__ = '0.1.0'

from .linkheader import Signposting,find_signposting
from .resolver import find_signposting_http
from httplink import Link

__all__ = [Signposting,find_signposting,find_signposting_http,Link]